
import ailib
from math import log, e, exp
from scipy import matrix

class Func0: # A * k^(-x) * log(l*c + c) + B
	"""
	       f(x) = A * k^-x * ln(l*x + c) + B
	 df(x) / dA = k^-x * ln(l*x + c)
	 df(x) / dk = A * -x * k^(x-1) * ln(l*x + c)
	 df(x) / dl = A * x * k^-x / (l * x + c)
	 df(x) / dc = A * k^-x / (l * x + c)
	 df(x) / dB = 1

	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * t[1,0]**(-x) * log(t[2,0] * x + t[3,0]) + t[4,0]
		dA = lambda t, x: t[1,0]**(-x) * log(t[2,0] * x + t[3,0])
		dk = lambda t, x: t[0,0] * -x * t[1,0]**(x-1) * log(t[2,0] * x + t[3,0])
		dl = lambda t, x: t[0,0] * x * t[1,0]**(-x) / (t[2,0] * x + t[3,0])
		dc = lambda t, x: t[0,0] * t[1,0]**(-x) / (t[2,0] * x + t[3,0])
		dB = lambda t, x: 1.0
		
		self.model = f
		self.partials = (dA,dk,dl,dc,dB)
		self.init = matrix([1.0, e, 1.0, 1.0, 1.0]).T
	
	def str(self, m):
		t = m.coeff
		return "%f * %f ** (-x) * log(%f * x + %f) + %f" % (t[0,0], t[1,0], t[2,0], t[3,0], t[4,0])

class Func1: # A * x * exp(-c * x) + B
	"""
	       f(x) = A * x * exp(-c0 * x) + B
	 df(x) / dA  = x * exp(-c0 * x)
	 df(x) / dc0 = - A * x**2 * exp(-c0 * x)
	 df(x) / dB  = 1

	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * x * exp(-x * t[1,0]) + t[2,0]
		dA = lambda t, x: x * exp(-t[1,0] * x)
		dc = lambda t, x: -t[0,0] * x*x * exp(-t[1,0] * x)
		dB = lambda t, x: 1.0
		
		self.model = f
		self.partials = (dA,dc,dB)
		self.init = matrix([1.0, 1.0, 0.0]).T
	
	def str(self, m):
		t = m.coeff
		return "%f * x * exp(-x) + %f" % (t[0,0], t[1,0])

class Func2: # A * log(x+1) * exp(-c*x) + B
	"""
	       f(x)  = A * log(x+1) * exp(-c*x) + B
	 df(x) / dA = log(x+1) * exp(-c*x)
	 df(x) / dc = -A * x * log(x+1) * exp(-c*x)
	 df(x) / dB = 1

	"""
	def __init__(self):
		f   = lambda t, x: t[0,0] * log(x+1.0) * exp(-t[1,0]*x) + t[2,0]
		dA  = lambda t, x: log(x+1) * exp(-t[1,0]*x)
		dc  = lambda t, x: -t[0,0] * x * log(x+1) * exp(-t[1,0]*x)
		dB  = lambda t, x: 1.0
		
		self.model = f
		self.partials = (dA,dc,dB)
		self.init = matrix([1.0, 0.5, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-%f * x) + %f" % (t[0,0], t[1,0], t[2,0])



class Func3: # A * log(x+1) * exp(-c*x^k) + B
	"""
	f(x) = A * log(x+1) * exp( -c * x**k ) + B
	d/dA = log(x+1) * exp( -c * x**k )
	d/dB = 1
	d/dc = -x**k * A * log(x+1) * exp(-c*x**k )
	d/dk = -c * x**k * ln(x) * A * log(x+1) * exp(-c*x**k)
	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * log(x+1) * exp( -t[1,0] * x**t[2,0] ) + t[3,0]
		dA = lambda t, x: log(x+1) * exp( -t[1,0] * x**t[2,0] )
		dB = lambda t, x: 1.0
		dc = lambda t, x: f(t,x) * -x**t[2,0]
		dk = lambda t, x: f(t,x) * -t[1,0] * x**t[2,0] * log(x+0.00000001)
		
		self.model = f
		self.partials = (dA,dc,dk,dB)
		self.init = matrix([1.0, 0.5, 2.0, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-%f * x**(%f)) + %f" % (t[0,0], t[1,0], t[2,0], t[3,0])

class Func4: # A * log(x+1) * exp(-c*(x-m)^k) + B
	"""
	f(x) = A * log(x+1) * exp( -c * (x-m)**k ) + B
	d/dA = log(x+1) * exp( -c * (x-m)**k )
	d/dB = 1
	d/dc = -(x-m)**k * A * log(x+1) * exp(-c*(x-m)**k )
	d/dm = c * k * (x-m)**(k-1) * A * log(x+1) * exp(-c*(x-m)**k)
	d/dk = -c * (x-m)**k * log(x-m) * A * log(x+1) * exp(-c*(x-m)**k)
	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * log(x+1) * exp( -t[1,0] * (x-t[2,0])**t[3,0] ) + t[4,0]
		dA = lambda t, x: log(x+1) * exp( -t[1,0] * (x-t[2,0])**t[3,0] )
		dB = lambda t, x: 1.0
		dc = lambda t, x: f(t,x) * -(x-t[2,0])**t[3,0]
		dm = lambda t, x: f(t,x) * t[1,0] * t[3,0] * (x-t[2,0])**(t[3,0]-1)
		dk = lambda t, x: f(t,x) * -t[1,0] * (x-t[2,0])**t[3,0] * log(x-t[2,0]+0.00000001)
		
		self.model = f
		self.partials = (dA,dc,dm,dk,dB)
		self.init = matrix([1.0, 0.5, 0.0, 2.0, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-%f * (x - %f)**(%f)) + %f" % (t[0,0], t[1,0], t[2,0], t[3,0], t[4,0])

class Func5: # A * log(x+1) * exp(-c*|x-m|^k) + B
	"""
	f(x) = A * log(x+1) * exp( -c * abs(x-m)**k ) + B
	d/dA = log(x+1) * exp( -c * abs(x-m)**k )
	d/dB = 1
	d/dc = A * log(x+1) * exp(-c*abs(x-m)**k ) * -abs(x-m)**k
	d/dm = A * log(x+1) * exp(-c*abs(x-m)**k ) * c * k * abs(x-m)**k / (x-m)
	d/dk = A * log(x+1) * exp(-c*abs(x-m)**k ) * c * ln(abs(x-m)) * -abs(x-m)**k
	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * log(x+1) * exp( -t[1,0] * abs(x-t[2,0])**t[3,0] ) + t[4,0]
		dA = lambda t, x: log(x+1) * exp( -t[1,0] * abs(x-t[2,0])**t[3,0] )
		dB = lambda t, x: 1.0
		dc = lambda t, x: f(t,x) * -abs(x-t[2,0])**t[3,0]
		dm = lambda t, x: f(t,x) * t[1,0] * t[3,0] * abs(x-t[2,0])**(t[3,0]-1) / (x-t[2,0] + 0.00000001)
		dk = lambda t, x: f(t,x) * -t[1,0] * abs(x-t[2,0])**t[3,0] * log(abs(x-t[2,0])+0.00000001)
		
		self.model = f
		self.partials = (dA,dc,dm,dk,dB)
		self.init = matrix([1.0, 0.5, 0.0, 2.0, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-%f * abs(x - %f)**(%f)) + %f" % (t[0,0], t[1,0], t[2,0], t[3,0], t[4,0])

class Func6: # A * log(x+1) * exp(-|c| * |x-m|^k) + B
	"""
	f(x) = A * log(x+1) * exp( -abs(c) * abs(x-m)**k ) + B
	d/dA = log(x+1) * exp( -abs(c) * abs(x-m)**k )
	d/dB = 1
	d/dc = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(x-m)**k * -sign(c)
	d/dm = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(c) * k * abs(x-m)**k / (x-m)
	d/dk = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(c) * ln(abs(x-m)) * -abs(x-m)**k
	"""
	def __init__(self):
		f  = lambda t, x: t[0,0] * log(x+1) * exp( -abs(t[1,0]) * abs(x-t[2,0])**t[3,0] ) + t[4,0]
		dA = lambda t, x: log(x+1) * exp( -abs(t[1,0]) * abs(x-t[2,0])**t[3,0] )
		dB = lambda t, x: 1.0
		dc = lambda t, x: f(t,x) * abs(x-t[2,0])**t[3,0] * ailib.sign(-t[1,0])
		dm = lambda t, x: f(t,x) * abs(t[1,0]) * t[3,0] * abs(x-t[2,0])**(t[3,0]-1) / (x-t[2,0] + 0.00000001)
		dk = lambda t, x: f(t,x) * -abs(t[1,0]) * abs(x-t[2,0])**t[3,0] * log(abs(x-t[2,0])+0.00000001)
		
		self.model = f
		self.partials = (dA,dc,dm,dk,dB)
		self.init = matrix([1.0, 0.5, 0.0, 2.0, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-abs(%f) * abs(x - %f)**(%f)) + %f" % (t[0,0], t[1,0], t[2,0], t[3,0], t[4,0])

#f(x) = A * log(s*x-o+1) * exp( -c * (s*x-m-o)**k ) + B
#f(x) = A * log(s*x+o+1) * exp(-c*(s*x+o)) + B # Func2 mit shift

class Func7: #Func2 beidseitig; B1=B2, Nullpkt
	"""
	       f(x)  = A * log(x+1) * exp(-c*x) + B
	 df(x) / dA = log(x+1) * exp(-c*x)
	 df(x) / dc = -A * x * log(x+1) * exp(-c*x)
	 df(x) / dB = 1

	beidseitig um t
		x <  t => f(-x)
		x >= t => f(+x)
	mit t als Ursprung
		x <  t => f(t-x)
		x >= t => f(x-t)
	=> f(abs(x-t)) aber andere params!
	
	params A,c twice
	"""
	def __init__(self):
		# t = matrix([a0, c0, a1, c1, b, t]).T
		def f(t, x):
			if x < t[5,0]:
				return t[0,0] * log(t[5,0]-x+1.0) * exp(-t[1,0]*(t[5,0]-x)) + t[4,0]
			else:
				return t[2,0] * log(x-t[5,0]+1.0) * exp(-t[3,0]*(x-t[5,0])) + t[4,0]
		dA  = lambda c, x: log(x+1) * exp(-c*x)
		dc  = lambda a, c, x: -a * x * log(x+1) * exp(-c*x)
		dA1  = lambda t, x: dA(t[0,0],+t[5,0] - x)
		dA2  = lambda t, x: dA(t[2,0],-t[5,0] + x)
		dc1  = lambda t, x: dc(t[0,0],t[1,0],+t[5,0] - x)
		dc2  = lambda t, x: dc(t[2,0],t[3,0],-t[5,0] + x)
		dB  = lambda t, x: 1.0
		def df(t, x):
			if x < t[5,0]:
				#-c * f(t,t-x) + a * exp(c*x-c*t) / (t-x+1) 
				return -t[1,0] * f(t,x-t[5,0]) + t[0,0] * exp(t[1,0]*(x-t[5,0])) / (t[5,0]-x+1.0)
			else:
				#+c * f(t,x-t) + a * exp(c*t-c*x) / (t-x-1)
				return t[3,0] * f(t,x-t[5,0]) + t[2,0] * exp(t[3,0]*(x-t[5,0])) / (t[5,0]-x-1.0)
		
		self.model = f
		self.partials = (dA1,dc1,dA2,dc2,dB, lambda t,x:0.0)
		self.init = matrix([1.0, 0.5, 1.0, 0.5, 0.0, 0.43]).T

	def str(self, m):
		t = m.coeff
		s  = "f(x) = %f * log(%f - x + 1.0) * exp(-%f * (%f - x)) + %f\n" % (t[5,0],t[0,0], t[5,0], t[1,0], t[5,0], t[4,0])
		s += "g(x) = %f * log(x - %f + 1.0) * exp(-%f * (x - %f)) + %f\n" % (t[5,0],t[2,0], t[5,0], t[3,0], t[5,0], t[4,0])
		s += "t = %f\n" % (t[5,0])






class Func2a: # Func2 with fixed B
	"""
	       f(x)  = A * log(x+1) * exp(-c*x) + B
	 df(x) / dA = log(x+1) * exp(-c*x)
	 df(x) / dc = -A * x * log(x+1) * exp(-c*x)
	 df(x) / dB = 1

	"""
	def __init__(self, B=0.377747):
		f   = lambda t, x: t[0,0] * log(x+1.0) * exp(-t[1,0]*x) + t[2,0]
		dA  = lambda t, x: log(x+1) * exp(-t[1,0]*x)
		dc  = lambda t, x: -t[0,0] * x * log(x+1) * exp(-t[1,0]*x)
		dB  = lambda t, x: 0.0
		
		self.model = f
		self.partials = (dA,dc,dB)
		self.init = matrix([1.0, 0.5, B]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+1) * exp(-%f * x) + %f" % (t[0,0], t[1,0], t[2,0])

class Func6a: # Func6, beidseitig
	"""
	f(x) = A * log(x+1) * exp( -abs(c) * abs(x-m)**k ) + B
	d/dA = log(x+1) * exp( -abs(c) * abs(x-m)**k )
	d/dB = 1
	d/dc = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(x-m)**k * -sign(c)
	d/dm = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(c) * k * abs(x-m)**k / (x-m)
	d/dk = A * log(x+1) * exp(-abs(c)*abs(x-m)**k ) * abs(c) * ln(abs(x-m)) * -abs(x-m)**k
	"""
	def __init__(self):
		# t = [A1,c1,m1,k1,A2,c2,m2,k2,B,t]
		f  = lambda t, x: t[0,0] * log(x+1) * exp( -abs(t[1,0]) * abs(x-t[2,0])**t[3,0] ) + t[4,0]
		def f(t, x):
			if x < t[9,0]:
				return t[0,0] * log(t[9,0]-x+1.0) * exp( -abs(t[1,0]) * abs(t[9,0]-x-t[2,0])**t[3,0] ) + t[8,0]
			else:
				return t[4,0] * log(x-t[9,0]+1.0) * exp( -abs(t[5,0]) * abs(x-t[9,0]-t[6,0])**t[7,0] ) + t[8,0]
		
		dA  = lambda x,c,m,k: log(x+1) * exp( -abs(c) * abs(x-m)**k )
		dc  = lambda x,c,m,k,t: f(t,x) * abs(x-m)**k * ailib.sign(-c)
		dm  = lambda x,c,m,k,t: f(t,x) * abs(c) * k * abs(x-m)**(k-1) / (x-m + 0.00000001)
		dk  = lambda x,c,m,k,t: f(t,x) * -abs(c) * abs(x-m)**k * log(abs(x-m)+0.00000001)
		dA1 = lambda t, x: dA( t[9,0] - x,t[1,0],t[2,0],t[3,0])
		dA2 = lambda t, x: dA(-t[9,0] + x,t[1,0],t[2,0],t[3,0])
		dc1 = lambda t, x: dc( t[9,0] - x,t[1,0],t[2,0],t[3,0], t)
		dc2 = lambda t, x: dc(-t[9,0] + x,t[5,0],t[6,0],t[7,0], t)
		dm1 = lambda t, x: dm( t[9,0] - x,t[1,0],t[2,0],t[3,0], t)
		dm2 = lambda t, x: dm(-t[9,0] + x,t[5,0],t[6,0],t[7,0], t)
		dk1 = lambda t, x: dk( t[9,0] - x,t[1,0],t[2,0],t[3,0], t)
		dk2 = lambda t, x: dk(-t[9,0] + x,t[5,0],t[6,0],t[7,0], t)
		dB  = lambda t, x: 1.0
		dt  = lambda t, x: 0.0
		
		self.model = f
		self.partials = (dA1,dc1,dm1,dk1,dA2,dc2,dm2,dk2,dB,dt)
		self.init = matrix([1.0, 0.5, 0.0, 2.0, 1.0, 0.5, 0.0, 2.0, 0.0, 0.43]).T

	def str(self, m):
		t = m.coeff
	
		s  = "f(x) = %f * log(%f-x+1) * exp(-abs(%f) * abs(%f-x - %f)**(%f)) + %f\n" % (t[0,0], t[9,0], t[1,0], t[9,0], t[2,0], t[3,0], t[8,0])
		s += "g(x) = %f * log(x-%f+1) * exp(-abs(%f) * abs(x-%f - %f)**(%f)) + %f\n" % (t[4,0], t[9,0], t[5,0], t[9,0], t[6,0], t[7,0], t[8,0])
		s += "t = %f\n" % (t[9,0])
		return s





class Func7a: #Func2 beidseitig; B1 != B2, Nullpkt
	def __init__(self):
		# t = matrix([a0, c0, a1, c1, b, t]).T
		def f(t, x):
			if x < t[5,0]:
				return t[0,0] * log(t[5,0]-x+1.0) * exp(-t[1,0]*(t[5,0]-x)) + t[4,0]
			else:
				return t[2,0] * log(x-t[5,0]+1.0) * exp(-t[3,0]*(x-t[5,0])) + t[6,0]
		import sys
		
		#dA  = lambda c, x: sys.stdout.write(str(x)+' '+str(c)+'\n') or log(x+1+0.000000001) * exp(-c*x)
		dA  = lambda c, x: log(x+1+0.000000001) * exp(-c*x)
		dc  = lambda a, c, x: -a * x * log(x+1) * exp(-c*x)
		dA1  = lambda t, x: dA(t[0,0],+t[5,0] - x)
		dA2  = lambda t, x: dA(t[2,0],-t[5,0] + x)
		dc1  = lambda t, x: dc(t[0,0],t[1,0],+t[5,0] - x)
		dc2  = lambda t, x: dc(t[2,0],t[3,0],-t[5,0] + x)
		dB  = lambda t, x: 1.0
		def df(t, x):
			if x < t[5,0]:
				#-c * f(t,t-x) + a * exp(c*x-c*t) / (t-x+1) 
				return -t[1,0] * f(t,x-t[5,0]) + t[0,0] * exp(t[1,0]*(x-t[5,0])) / (t[5,0]-x+1.0)
			else:
				#+c * f(t,x-t) + a * exp(c*t-c*x) / (t-x-1)
				return t[3,0] * f(t,x-t[5,0]) + t[2,0] * exp(t[3,0]*(x-t[5,0])) / (t[5,0]-x-1.0)
		
		def dAs(t,x):
			if x < t[5,0]:
				return dA(t[0,0],+t[5,0] - x)
			else:
				return dA(t[2,0],-t[5,0] + x)
		
		def dcs(t,x):
			if x < t[5,0]:
				return dc(t[0,0],t[1,0],+t[5,0] - x)
			else:
				return dc(t[2,0],t[3,0],-t[5,0] + x)
		
		self.model = f
		#self.partials = (dA1,dc1,dA2,dc2,dB, lambda t,x:0.0, dB)
		self.partials = (dAs,dcs,lambda t,x:0.0,lambda t,x:0.0,dB, lambda t,x:0.0, dB)
		self.init = matrix([1.0, 0.5, 1.0, 0.5, 0.0, 0.86, 0.0]).T


	def str(self, m):
		t = m.coeff
		s  = "f(x) = %f * log(%f - x + 1.0) * exp(-%f * (%f - x)) + %f\n" % (t[0,0], t[5,0], t[1,0], t[5,0], t[4,0])
		s += "g(x) = %f * log(x - %f + 1.0) * exp(-%f * (x - %f)) + %f\n" % (t[2,0], t[5,0], t[3,0], t[5,0], t[6,0])
		s += "t = %f\n" % (t[5,0])
		return s

class Func2b: # Func2 with t
	"""
	       f(x)  = A * log(x+t+1) * exp(-c*(x+t)) + B
	 df(x) / dA = log(x+t+1) * exp(-c*(x+t))
	 df(x) / dc = -A * (x+t) * log(x+t+1) * exp(-c*(x+t))
	 df(x) / dB = 1
	 df(x) / dt = A * exp(-c*(x+t)) * (1/(x+t+1) - c*log(x+t+1))

	"""
	def __init__(self):
		f   = lambda t, x: t[0,0] * log(x+t[3,0]+1.0) * exp(-t[1,0]*(x+t[3,0])) + t[2,0]
		dA  = lambda t, x: log(x+t[3,0]+1.0) * exp(-t[1,0]*(x+t[3,0]))
		dc  = lambda t, x: -t[0,0] * (x+t[3,0]) * log(x+t[3,0]+1.0) * exp(-t[1,0]*(x+t[3,0]))
		dB  = lambda t, x: 1.0
		dt  = lambda t, x: t[0,0] * exp(-t[1,0]*(x+t[3,0])) * ( 1.0/(x+t[3,0]+1.0) - t[1,0]*log(x+t[3,0]+1.0) )
		#dt  = lambda t, x: 0.0 # Debug: Don't change t
		
		self.model = f
		self.partials = (dA,dc,dB,dt)
		self.init = matrix([1.0, 0.5, 0.0, 0.0]).T

	def str(self, m):
		t = m.coeff
		return "%f * log(x+%f+1) * exp(-%f * (x+%f)) + %f" % (t[0,0], t[3,0], t[1,0], t[3,0], t[2,0])
