import Image
import ImageChops
import sys
import cv

"""
detect:
    1. Get frame
    2. Compute diff to last frame
    3. If diff > thres  => motion
    4. Otherwise        => no motion

detection sequence:
0. no motion (init)
1. motion
2. no motion
    -> output 'Next page'

threshold high => motion-part may be fragmented
threshold low  => no-motion-part may be fragmented
=> majority vote on last N frames
=> 2-layer: majority vote on last M votes

To plot, use
    gnuplot> plot __motionfile__ with lines, __distfile__ with lines, __thres__

The following output may occur (but be ignored): 'VIDIOC_QUERYMENU: Invalid argument'
"""


class CaptureSource:
    """
    """
    def __init__(self): pass
    def grab(self): abstract()

class CaptureCV(CaptureSource):
    """
    """
    def __init__(self, device=0, width=640, height=480, skip=0, timeout=0):
        """
        """
        self.timeout = timeout >= 0 and timeout or 0
        self.width, self.height = width, height
        self.capture = cv.CreateCameraCapture(device)
        if not self.capture: raise Exception("Error opening capture device")
        
        cv.SetCaptureProperty(self.capture, cv.CV_CAP_PROP_FRAME_WIDTH, width)
        cv.SetCaptureProperty(self.capture, cv.CV_CAP_PROP_FRAME_HEIGHT, height)
        
        # skip some frames
        for i in range(skip): cv.QueryFrame(self.capture)

    def grab(self):
        """
        """
        for i in range(1 + self.timeout): self.last = cv.QueryFrame(self.capture)
        return Image.fromstring("L", (self.width, self.height), self.last.tostring())

class MotionTrigger:
    """
    """
    def __init__(self, capture, callback, windowSize=5, thres=None):
        """
        """
        self.capture = capture
        self.windowSize = windowSize
        self.thres = thres
        self.callback = callback
        self.dist = MotionTrigger.dist_sqr
        self.running = False
    
    def run(self, maxframe=-1):
        """
        
        :param maxframe:        Number of iterations. Negative for infinite.
        :type maxframe:         int
        """
        mhist = [0 for i in range(self.windowSize)]
        motion = False
        frame_no, trigger_no = 0, 0
        img_last, img_curr = None, self.capture.grab()
        trigger = False
        self.running = True
        
        # ensure threshold is set; auto adjust if necessary
        if self.thres is None:
            #sys.stdout.write('Auto adjusting threshold... ')
            #sys.stdout.flush()
            self.thres = self.autothres()
            #print "set to", self.thres
        
        while self.running and not (maxframe >= 0 and frame_no >= maxframe):

            # grab frame and compute image difference
            img_last = img_curr
            img_curr = self.capture.grab()
            diff = self.dist(img_curr, img_last)
            
            # determine motion or not
            if diff > self.thres:
                mhist.append(1)
            else:
                mhist.append(0)
            
            mhist.pop(0)
            if float(sum(mhist))/len(mhist) > 0.5:
                motion = True
            else:
                if motion: # change from motion to no motion -> trigger event
                    self.callback(frame_no, diff, motion)
                    trigger = True
                motion = False
            
            # iteration cleanup
            self._post(frame_no, img_curr, diff, motion, trigger)
            frame_no += 1
            trigger = False

    def autothres(self, numLoops=20, scale=1.5):
        thres = 0
        img_last, img_curr = None, self.capture.grab()
        for i in range(numLoops):
            img_last = img_curr
            img_curr = self.capture.grab()
            thres += self.dist(img_curr, img_last)
        
        # use scaled mean of sampled distances
        return (thres * scale) / numLoops

    def _post(self, frm, curr, diff, motion, trigger): pass

    @staticmethod
    def dist_diff(curr, prev): # curr - prev
        c = curr.load()
        p = prev.load()
        W,H = curr.size
        d = sum([c[x,y] - p[x,y] for y in range(H) for x in range(W)])
        return d

    @staticmethod
    def dist_abs(curr, prev): # | curr - prev |
        diff = ImageChops.difference(curr, prev).load()
        d = sum([diff[x,y] for y in range(curr.size[1]) for x in range(curr.size[0])])
        return d

    @staticmethod
    def dist_sqr(curr, prev): # (curr - prev)**2
        c = curr.load()
        p = prev.load()
        W,H = curr.size
        d = sum([(c[x,y] - p[x,y])**2 for y in range(H) for x in range(W)])
        return d





