import sys, include, re
from math import ceil, log
import modRansac
from Matrix33 import Matrix33
from Shape import Point, Square
from custImage import custImage

"""
 Ransac()
 > Input
 > Calculate matrix => Matrix33
 > Matrix33.transform(Point)

Ransac:
	input Vectors (src, dst)
	Algorithm parameters: loops, setSize, threshold
	
	transMatrix, precision
	
	Fx: Scale
	Ox: Offset - Position der Referenz
	
	initCalc(e: Double): set loops, setSize
	calcPoint(p: Point): return p' = M*p
	invert(m: Matrix33): 
	

"""
class Ransac:
	# input matrix
	# loops, setSize, threshold
	# transMatrix
	# precision
	# calcPoint
	# invertMatrix
	# fx, ox (ox = Position der Referenz!)
	def __init__(self, threshold=10, inVec=None):
		self.loops = 7
		self.setSize = 10
		self.threshold = threshold
		self.transMatrix = Matrix33()
		self.scale = None
		self.offset = None
		self.__scaleMethod = self.__scaleMax
		self.inverted = None

	def initParams(self, e, n=0):
		if n == 0: n = len(self.inVec)
		p = 0.99
		self.setSize = ceil((1 - e) * n)
		self.loops = ceil( log( 1 - p ) / log( 1-pow(1-e, 4) ) )

	def setParams(self, setSize, loops):
		self.setSize = setSize
		self.loops = loopy

	# execute ransac, store in self.matrix, return self.precision
	def calcMatrix(self, inVec=None):
		if inVec == None:	inVec = self.inVec
		else:			self.inVec = inVec
		
		if len(inVec) < 4: raise Exception('At least 4 Point pairs are required')
		cVec = map(lambda (a,b): ([a.x, a.y, 1.0], [b.x, b.y, 1.0]), inVec)
		self.precision, self.transMatrix.matrix = modRansac.ransac(cVec, self.setSize, self.loops, 0.0, self.threshold)
		return self.precision

	def plainHomography(self, inVec=None):
		if inVec == None:	inVec = self.inVec
		else:				self.inVec = inVec
		
		if len(inVec) < 4: raise Exception('At least 4 Point pairs are required')
		cVec = map(lambda (a,b): ([a.x, a.y, 1.0], [b.x, b.y, 1.0]), inVec)
		self.transMatrix.matrix = modRansac.getHomography(cVec)
		self.precision = 1.0
		return self.precision

	# calc self.matrix * point
	def calcPoint(self, point, matrix=None):
		if matrix == None: matrix = self.transMatrix
		# use python implementation
		#return matrix.transform(point)
		
		# use c implementation
		pt = modRansac.transform(matrix.matrix, [point.x, point.y, 1.0])
		return Point(pt[0], pt[1])

	# invert self.matrix
	def invertMatrix(self):
		if self.inverted != None: return self.inverted

		# use python implementation
		#self.inverted = self.transMatrix.invert()
		#return self.inverted
		
		# use c implementation
		self.inverted = Matrix33()
		self.inverted.matrix = modRansac.invert(self.transMatrix.matrix)
		return self.inverted

	"""
	Origin: Matrix A, Scale S = (fx, fy), Image dimension R = (width, height)
	p0' = A^(-1)*(0, 0)
	p1' = A^(-1)*(Rw, 0)
	p2' = A^(-1)*(0, Rh)
	p3' = A^(-1)*(Rw, Rh)
	ox = Sx * min(p0'.x, p1'.x, p2'.x, p3'.x)
	oy = Sy * min(p0'.y, p1'.y, p2'.y, p3'.y)
	"""
	# requires getScale (scale, transMatrix)
	def getOffset(self, image, scale=None):
		if scale == None: scale = self.scale
		m = self.invertMatrix()
		width, height = float(image.width), float(image.height)
		pts = map(lambda x: self.calcPoint(Point(x[0], x[1]), m), [(0, 0), (width, 0), (0, height), (width, height)])
		self.offset = Point(abs(scale[0] * min(map(lambda p: p.x, pts))), abs(scale[1] * min(map(lambda p: p.y, pts))))
		return self.offset

	"""
	Scale: Matrix A, Reference Dimension R = (width, height)
	p12 = A*(0, 0) - A*(Rw, 0)
	p13 = A*(0, 0) - A*(0, Rh)
	p43 = A*(Rw, Rh) - A*(0, Rh)
	p42 = A*(Rw, Rh) - A*(Rw, 0)
	
	w' = F(p12, p43), h' = F(p13, p42)
	w'**2 * (Rh/Rw) > h'**2 * (Rw/Rh):	h' = w' * Rh/Rw
	else:					w' = h' * Rw/Rh

	fx = w'/Rw
	fy = h'/Rh
	"""
	# require calcMatrix (transMatrix)
	def getScale(self, image):
		width, height = float(image.width), float(image.height)
		pts = map(lambda x: self.calcPoint(Point(x[0], x[1])), [(0, 0), (width, 0), (0, height), (width, height)])
		w = float(self.__scaleMethod(pts[0] - pts[1], pts[3] - pts[2]))
		h = float(self.__scaleMethod(pts[0] - pts[2], pts[3] - pts[1]))
		
		if w**2 * (height / width) > h**2 * (width / height):	h = w * (height / width)
		else:							w = h * (width / height)
		
		self.scale = (w / width, h / height)
		return self.scale

	def __scaleMax(self, fst, snd):		return max(fst.norm(), snd.norm())
	def __scaleMin(self, fst, snd):		return min(fst.norm(), snd.norm())
	def __scaleAvg(self, fst, snd):		return (fst.norm() + snd.norm())/2.0

	def toTransformation(self, mode='MATRIX'):
		trans = Transformation()
		if mode == 'MATRIX':
			trans.initByMatrix(self.transMatrix)
		elif mode == 'SQUARE':
			trans.initBySquare(Point(self.offset[0], self.offset[1]), self.scale)
		elif mode == 'SINGLE':
			trans.initBySingle(Point(self.offset[0], self.offset[1]))
		return trans

class App:
	def __init__(self):
		self.outIm = False
		self.output = sys.stdout
		self.e = 0.0
		self.imgPath, self.refPath = '', ''
		self.img, self.ref = None, None
	
	def run(self):
		self.parseArgs(sys.argv[1:])
		self.initialize()
		self.execute()
		self.printOut()
		return self
	
	def parseArgs(self, argv):
		import oParse
		
		try:
			opts, self.rest = oParse.getopt(argv, ['l=', 's=', 'e=', 'im=', 'f=', 'i=', 'o=', 't='], ['ref=', 'img=', 'input=', 'output=', 'threshold=', 'loops=', 'setsize='])
			for opt, arg in opts:
				if opt in ('-f', '-i', '--input'):
					self.inPath = arg
				elif opt in ('-t', '--threshold'):
					self.threshold = float(arg)
				elif opt in ('-l', '--loops'):
					self.loops = int(arg)
				elif opt in ('-s', '--setsize'):
					self.setSize = int(arg)
				elif opt in ('-e'):
					self.e = float(arg)
				elif opt in ('-im'):
					self.outIm = True
				elif opt in ('-o', '--output'):
					self.output = open(arg, 'w')
				elif opt in ('--img'):
					self.imgPath = arg
				elif opt in ('--ref'):
					self.refPath = arg
			
		except oParse.GetoptError, e:
			sys.stderr.write("An error was detected while parsing parameter" + str(e) + '\n')
			self.helpAndExit(1)
		
		return self

	def readInput(self, fileName):
		vec = []
		f = open(fileName, 'r')
		for line in f:
			m = re.match('\((\d+(?:\.\d+)?)\s*,\s*(\d+(?:\.\d+)?)\)\s*\((\d+(?:\.\d+)?)\s*,\s*(\d+(?:\.\d+)?)\)', line)
			if m != None:
				x0, x1, y0, y1 = map(lambda x: float(x), m.groups())
			vec.append((Point(x0, x1), Point(y0, y1)))
		return vec

	def initialize(self):
		self.inVec = self.readInput(self.inPath)
		self.ransac = Ransac(self.threshold)
		if self.e != 0.0:
			self.ransac.initCalc(self.e)
		else:
			self.ransac.loops = self.loops
			self.ransac.setSize = self.setSize
		self.ransac.inVec = self.inVec
		if self.refPath != '':
			self.ref = custImage(self.refPath)
			if self.imgPath != '': self.img = custImage(self.imgPath)
		return self
	
	def execute(self):
		self.ransac.calcMatrix()
		if self.ref != None:
			self.ransac.getScale(self.ref)
			if self.img != None:
				self.ransac.getOffset(self.img)
		return self
	
	def printOut(self):
		if self.outIm:
			pass
		else:
			self.output.write(str(self.ransac.transMatrix) + '\n')
			if self.ransac.scale != None: self.output.write('(fx, fy) = ' + str(self.ransac.scale) + '\n')
			if self.ransac.offset != None: self.output.write('(ox, oy) = ' + str(self.ransac.offset) + '\n')
		return self
	
	def helpAndExit(self, code=0):
		exit(code)


if __name__ == '__main__': App().run()



