
import sys, include
import Image, ImageDraw, random, time


class randomImage:
	# mode: '1' for purely black/white
	#	'L' for grayscale
	#	'RGB' for color images (in this case, colors/background need to be triples)
	def __init__(self, mode='1', colors=[0], background=255, blkProb=0.5, sigmaFac=0.25, randomize=False):
		self.mode = mode
		self.colors = colors
		self.background = background
		self.blkProb = blkProb
		self.sigmaFac = sigmaFac
		self.randomize = randomize

	def fromConfig(self, cfg):
		for key in ('mode', 'colors', 'background', 'blkProb', 'sigmaFac', 'randomize'):
			try: self.__dict__[key] = cfg[key]
			except KeyError: pass

	def createImage(self, imgSize, blkSize, randomize=False):
		if randomize or self.randomize:	blkFunc = self.__blkRandomized
		else:				blkFunc = self.__blkAligned

		# init
		img = Image.new(self.mode, imgSize, self.background)
		draw = ImageDraw.Draw(img)
		
		# fill with rectangles
		for x in range(0, imgSize[0], blkSize[0]):
			for y in range(0, imgSize[1], blkSize[1]):
				if random.randint(0, 99) < 100 * self.blkProb:
					blkFunc(draw, (x,y), blkSize)
		
		# return
		return img

	def __blkAligned(self, d, (x,y), blkSize):
		d.rectangle([(x,y), (x+blkSize[0], y+blkSize[1])], fill=random.choice(self.colors))

	def __blkRandomized(self, d, (x,y), blkSize):
		width, height = [random.gauss(blkSize[i], blkSize[i] * self.sigmaFac) for i in (0,1)]
		d.rectangle([(x,y), (x+width, y+height)], fill=random.choice(self.colors))


class App:
	def __init__(self):
		self.output = sys.stdout
		self.seed = 0
		self.verbose = False
		self.dstPath = ''
		self.randomize = False
		self.imgCreator = randomImage()

	def run(self):
		self.parseArgs(sys.argv[1:])
		self.initialize()
		self.execute()
		self.printOut()
		return self
	
	def parseArgs(self, argv):
		import oParse
		
		try:
			opts, self.rest = oParse.getopt(argv, ['s=', 'b=', 'd=', 'o=', 'r', 'v', 'h'], ['size=', 'seed=', 'block=', 'dst=', 'output=', 'randomize=', 'verbose', 'help'])
			for opt, arg in opts:
				if opt in ('-s', '--size'):
					self.imgSize = map(int, arg.split('x'))
				elif opt in ('-b', '--block'):
					self.blkSize = map(int, arg.split('x'))
					if len(self.blkSize) == 1: self.blkSize = (self.blkSize[0], self.blkSize[0])
				elif opt in ('-r', '--randomize'):
					self.randomize = True
					if opt == '--randomize': self.imgCreator.sigmaFac = float(arg)
				elif opt in ('--seed'):
					self.seed = int(arg)
				elif opt in ('-d', '--dst', '-o', '--output'):
					self.dstPath = arg
				elif opt in ('-v', '--verbose'):
					self.verbose = True
				elif opt in ('-h', '--help'):
					self.helpAndExit()
			if len(self.rest) == 1 and self.dstPath == '': self.dstPath = self.rest[0]
			
		except oParse.GetoptError, e:
			sys.stderr.write("An error was detected while parsing parameter" + str(e) + '\n')
			self.helpAndExit(1)
		
		return self
	
	def initialize(self):
		# set random seed
		if self.seed == 0:
			t = time.time()
			self.seed = int((t-int(t))*1000000)
		random.seed(self.seed)
		
		if self.verbose: print "Generating image with seed = ", self.seed
		return self
	
	def execute(self):
		self.last = self.imgCreator.createImage(self.imgSize, self.blkSize, self.randomize)
		return self
	
	def printOut(self):
		try:
			self.last.save(self.dstPath)
		except KeyError:
			self.last.save(self.dstPath, 'PNG')
		return self
	
	def helpAndExit(self, code=0):
		print "Create an image with nothing but randomly placed blocks"
		print ""
		print "Usage: <cmd> [params] <target file>"
		print "-s --size		Image size [wxh]"
		print "-b --block		Block dimension [wxh]"
		print "-r --randomize		Randomize block dimension"
		print "			(--randomize [Sigma])"
		print "-d --dst"
		print "-o --output		Output file"
		print "--seed			Set forced seed"
		print "-v --verbose		Some output"
		print "-h --help		This Help"
		exit(code)


if __name__ == '__main__': App().run()



