
#include <Python.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <string.h>

#include "detect/imload.h"
#include "detect/surflib.h"

#include "detect/os_mapping.h"

using namespace std;
using namespace surf;

// Length of the descriptor vector
int VLength;


vector< Ipoint > doDetect(char* inFile, bool doubleImageSize) {
	// Initial sampling step (default 2)
	int samplingStep = 2;
	// Number of analysed octaves (default 4)
	int octaves = 4;
	// Blob response treshold
	double thres = 4.0;
	// Set this flag "true" to double the image size
	// bool doubleImageSize = false;
	// Initial lobe size, default 3 and 5 (with double image size)
	int initLobe = 3;
	// Upright SURF or rotation invaraiant
	bool upright = false;
	// If the extended flag is turned on, SURF 128 is used
	bool extended = true;
	// Spatial size of the descriptor window (default 4)
	int indexSize = 4;

	ImLoad ImageLoader;
	Image *im=NULL;
	im = ImageLoader.readImage(inFile); 

	// Create the integral image
	Image iimage(im, doubleImageSize);

	// These are the interest points
	vector< Ipoint > ipts;
	ipts.reserve(1000);

	// Extract interest points with Fast-Hessian
	FastHessian fh(&iimage, /* pointer to integral image */
		ipts,
		thres, /* blob response threshold */
		doubleImageSize, /* double image size flag */
		initLobe * 3 /* 3 times lobe size equals the mask size */, 
		samplingStep, /* subsample the blob response map */
		octaves /* number of octaves to be analysed */);

		// Extract them and get their pointer
		fh.getInterestPoints();

	// Initialise the SURF descriptor
	Surf des(&iimage, /* pointer to integral image */	
		doubleImageSize, /* double image size flag */ 
		upright, /* rotation invariance or upright */
		extended, /* use the extended descriptor */
		indexSize /* square size of the descriptor window (default 4x4)*/);

	// Get the length of the descriptor vector resulting from the parameters
	VLength = des.getVectLength();

	// Compute the orientation and the descriptor for every interest point
	for (unsigned n=0; n<ipts.size(); n++){
		//for (Ipoint *k = ipts; k != NULL; k = k->next){
		// set the current interest point
		des.setIpoint(&ipts[n]);
		// assign reproducible orientation
		des.assignOrientation();
		// make the SURF descriptor
		des.makeDescriptor();
	}

	delete im;
	return ipts;
}

static PyObject* improve_detect(PyObject *self, PyObject *args) {
	char *inFile;
	bool doubleImageSize = false;
	if (! PyArg_ParseTuple( args, "s|b", &inFile, &doubleImageSize)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}

	vector< Ipoint > ipts = doDetect(inFile, doubleImageSize);

	// save the interest points in the output file
	PyObject *finList = PyList_New(0);
	for (unsigned n=0; n<ipts.size(); n++){
		PyObject *vecList = PyList_New(0);

		// circular regions with diameter 5 x scale
		double sc = 2.5 * ipts[n].scale; sc*=sc;
		PyList_Append(vecList, PyFloat_FromDouble(ipts[n].x));
		PyList_Append(vecList, PyFloat_FromDouble(ipts[n].y));
		PyList_Append(vecList, PyFloat_FromDouble(1.0/sc));
		PyList_Append(vecList, PyFloat_FromDouble(0.0));
		PyList_Append(vecList, PyFloat_FromDouble(1.0/sc));
		PyList_Append(vecList, PyInt_FromLong((long)ipts[n].laplace));
		PyList_Append(vecList, PyFloat_FromDouble(ipts[n].strength));
		PyList_Append(vecList, PyFloat_FromDouble(ipts[n].ori));

		// Here comes the descriptor
		for (int i = 0; i < VLength; i++) {
			PyList_Append(vecList, PyFloat_FromDouble(ipts[n].ivec[i]));
		}
		
		PyList_Append(finList, vecList);
	}
	
	return finList;
}






///*

static PyMethodDef DetectMethods[] = {
	{"detect",  improve_detect, METH_VARARGS, "Some temp module"},
	{NULL, NULL, 0, NULL}        // Sentinel
};

PyMODINIT_FUNC initmodDetect(void) {
	(void) Py_InitModule("modDetect", DetectMethods);
}

// */
