
#include "polynom.h"
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>


/**
 * X = Stützstellen
 * Y = Werte der Stützstellen
 * N = Anzahl Stützstellen (x, y Paare)
 * K = Polynomgrad
 * 
 * X = [ x1 x2 x3 ... ]^T ([N, 1])
 * Y = [ y1 y2 y3 ... ]^T ([N, 1])
 * M = [
 *      1  x1 x1^2 x1^3 .. x1^k
 *      1  x2 x2^2 x2^3 .. x2^k
 *      1  x3 x3^2 x3^3 .. x3^k
 *      .. 
 *     ] ([N, K+1])
 * P = ((M^T * M)^-1 * M^T) * Y
 * P = [ p1 p2 p3 p4 ... ]^T ([K+1, 1])
 * 
 * Polynom = p1 * x^0 + p2 * x^1 + p3 * x^2 + ...
 * 
 * 
 * @param X		X-Points
 * @param img		Y-Points
 * @param N		Dimension of X, img
 * @param K		Polynom degree
 * @param polynom	Target array. Needs to be allocated with (K+1) Elements
 *			if polynom == NULL, it will not be written !!
 *			otherwise, it is: polynom[0] + polynom[1] * x + polynom[2] * x^2 + polynom[3] * x^3 + ... + polynom[K] * x^K
 * @return		gsl vector with the same content as polynom
*/

gsl_vector* poly(double X[], double img[], int N, int K, double polynom[]) {
	
	// essential matrices
	gsl_vector *Y = gsl_vector_alloc(N);
	gsl_vector *P = gsl_vector_alloc(K+1);
	gsl_matrix *M = gsl_matrix_alloc(N, K+1);

	// work matrices
	gsl_matrix *Mt = gsl_matrix_alloc(K+1, N);
	gsl_matrix *MtM = gsl_matrix_alloc(K+1, K+1);
	gsl_matrix *inverse = gsl_matrix_alloc(K+1, K+1);
	gsl_matrix *fin = gsl_matrix_alloc(K+1, N);

	// work variables
	int i, k;
	double last, x;
	
	// Y
	for(i=0; i<N; i++) gsl_vector_set(Y, i, img[i]);
	
	// M
	for(i=0; i<N; i++) {
		gsl_matrix_set(M, i, 0, 1.0);
		x = X[i];
		last = 1.0;
		for(k=1; k<=K; k++) {
			
			last = last * x;
			gsl_matrix_set(M, i, k, last);
		}
	}
	
	// Mt
	gsl_matrix_transpose_memcpy(Mt, M);
	
	// P = ((M^T * M)^-1 * M^T) * Y
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, Mt, M, 0.0, MtM);	//MtM = Mt * M
	inverse = gsl_invert(MtM);						// inverse = MtM ^(-1)
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, inverse, Mt, 0.0, fin);	//fin = inverse * Mt
	gsl_blas_dgemv(CblasNoTrans, 1.0, fin, Y, 0.0, P);			//P = fin * Y
	
	// return P
	if (polynom != NULL) {
		for(i=0; i<=K; i++) {
			polynom[i] = gsl_vector_get(P, i);
		}
	}
	return P;
}



/**
 * @todo	Replace LU-invert by matrix calculations
 */
gsl_matrix* gsl_invert(gsl_matrix *In) {
	gsl_matrix *inverse, *m;
	gsl_permutation *p;
	int sig;
	
	if (In->size1 != In->size2) {
		printf("Cannot invert non-quadratic matrix\n");
		return NULL;
	}
	
	// copy matrix
	m = gsl_matrix_alloc(In->size1, In->size2);
	gsl_matrix_memcpy(m, In);
	
	// LU
	p = gsl_permutation_alloc(m->size1);
	gsl_linalg_LU_decomp (m, p, &sig);
	
	// inverse
	inverse = gsl_matrix_alloc(m->size1, m->size1);
	gsl_linalg_LU_invert (m, p, inverse);
	
	// free resources
	gsl_permutation_free(p);
	gsl_matrix_free(m);
	return inverse;
}



void gslPrintVector(gsl_vector* v, char *name) {
	int i;
	
	printf("%s = [ ", name);
	for(i=0; i<v->size; i++) {
		printf("%f ", gsl_vector_get(v, i));
	}
	printf("]\n");
}

void gslPrintMatrix(gsl_matrix *m, char *name) {
	int i, k, imax = m->size1, kmax = m->size2;
	
	printf("Matrix %s =\n", name);
	for(i=0; i<imax; i++) {
		printf("[ ");
		for(k=0; k<kmax; k++) {
			printf("%f ", gsl_matrix_get(m, i, k));
		}
		printf("]\n");
	}
}

