
#include <Python.h>

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "ransac/ransac.h"
#include "ransac/model.h"

#include <opencv/cv.h>

double vecNorm(vec d) {
	return sqrt(pow(d[0], 2) + pow(d[1], 2));
}

void vecDiff(vec a, vec b, vec d) {
	d[0] = a[0] - b[0];
	d[1] = a[1] - b[1];
	d[2] = 1;
}

void getPoint(matrix model, double x, double y, vec cpoint) {
	vec ipoint;
	ipoint[0] = x; ipoint[1] = y; ipoint[2] = 1;
	calcPoint(model, ipoint, cpoint);
}

double min(double a, double b) {
	if (a < b) return a;
	return b;
}

double max(double a, double b) {
	if (a < b) return b;
	return a;
}

PyObject* buildPoint(double x0, double y0, double x1, double y1) {
	PyObject *pt = PyTuple_New(2);
	PyObject *fst = PyTuple_New(2);
	PyObject *snd = PyTuple_New(2);
	
	PyTuple_SetItem(fst, 0, PyFloat_FromDouble(x0));
	PyTuple_SetItem(fst, 1, PyFloat_FromDouble(y0));
	PyTuple_SetItem(snd, 0, PyFloat_FromDouble(x1));
	PyTuple_SetItem(snd, 1, PyFloat_FromDouble(y1));
	
	PyTuple_SetItem(pt, 0, fst);
	PyTuple_SetItem(pt, 1, snd);
	return pt;
}



double doRansac(matrix model, vec *src, vec *dst, int dim, int set, int loops, double e, double dist) {

	r_data data = {src, dst, dim};
	r_param params = {set, loops, e, dist};
	if (params.e != 0) calcParams(&params, params.e, data.dim, MODEL_POINTS);

	double status;

	if (data.dim <= 0) return -1;
	if (params.dist <= 0) return -1;
	if (params.set <= 0 || params.loops <= 0) return -1;
	
	status = ransac(&data, &params, MODEL_POINTS, model);
	return status;
}

// improve.ransac(fst: [[Int]], snd: [[Int]], [set, loops, e, dist]
static PyObject* improve_ransac(PyObject *self, PyObject *args) {
	matrix model;
	int set=7, loops=10;
	double status, e=0.0, dist=10.0;
	
	// Get the arguments
	PyObject *mList;
	if (! PyArg_ParseTuple( args, "O!|iidd", &PyList_Type, &mList, &set, &loops, &e, &dist)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}

	// init lists
	int i, dim = PyList_Size(mList);
	vec *src = (vec*)malloc(sizeof(vec) * dim);
	vec *dst = (vec*)malloc(sizeof(vec) * dim);
	for (i=0; i<dim; i++){
		PyObject *srcList, *dstList, *tuple;
		
		// init srcList, dstList
		tuple = PyList_GetItem(mList, i);
		srcList = PyTuple_GetItem(tuple, 0);
		dstList = PyTuple_GetItem(tuple, 1);
		
		src[i][0] = PyFloat_AsDouble(PyList_GetItem(srcList, 0));
		src[i][1] = PyFloat_AsDouble(PyList_GetItem(srcList, 1));
		if (PyList_Size(srcList) >= 3) {
			src[i][2] = PyFloat_AsDouble(PyList_GetItem(srcList, 2)); // only, if given
		} else {
			src[i][2] = 1.0;
		}

		dst[i][0] = PyFloat_AsDouble(PyList_GetItem(dstList, 0));
		dst[i][1] = PyFloat_AsDouble(PyList_GetItem(dstList, 1));
		if (PyList_Size(dstList) >= 3) {
			dst[i][2] = PyFloat_AsDouble(PyList_GetItem(dstList, 2)); // only, if given
		} else {
			dst[i][2] = 1.0;
		}
		
		// printf("%2.5f\t%2.5f\t%2.5f\t%2.5f\n", src[i][0], src[i][1], dst[i][0], dst[i][1]);
	}

	// do the calculations
	//printf("set = %i, loops = %i, e = %f, dist = %f\n", set, loops, e, dist);
	//printf("src, dst; %i\n", dim);
	//for(i=0; i<dim; i++) {
	//	printf("(%f, %f, %f),\t(%f, %f, %f)\n", src[i][0], src[i][1], src[i][2], dst[i][0], dst[i][1], dst[i][2]);
	//}
	
	status = doRansac(model, src, dst, dim, set, loops, e, dist);
	
	// return
	if (status < 0) {
		//printf("set = %i, loops = %i, e = %f, dist = %f => status = %f\n", set, loops, e, dist, status);
		PyErr_SetString(PyExc_RuntimeError, "ransac calculating error");
		return NULL;
	}
	
	PyObject *list = PyList_New(0);
	int k;
	for (i=0; i<DIM; i++) {
		PyObject *sublist = PyList_New(0);
		for(k=0; k<DIM; k++) {
			//printf("%2.5f ", model[i][k]);
			PyList_Append(sublist, PyFloat_FromDouble(model[i][k]));
		}
		//printf("\n");
		PyList_Append(list, sublist);
	}
	
	free(src); free(dst);

	PyObject *finTuple = PyTuple_New(2);
	PyTuple_SetItem(finTuple, 0, PyFloat_FromDouble(status));
	PyTuple_SetItem(finTuple, 1, list);
	return finTuple;
	//return list;
}


int getModelFromPython(PyObject *mList, matrix model) {
	if (PyList_Size(mList) != 3) {
		PyErr_SetString(PyExc_RuntimeError, "Matrix dimension failure");
		return 1;
	}
	int i;
	for (i=0; i<3; i++){
		PyObject *colList = PyList_GetItem(mList, i);
		if (PyList_Size(colList) != 3) {
			PyErr_SetString(PyExc_RuntimeError, "Matrix dimension failure");
			return 1;
		}
		
		model[i][0] = PyFloat_AsDouble(PyList_GetItem(colList, 0));
		model[i][1] = PyFloat_AsDouble(PyList_GetItem(colList, 1));
		model[i][2] = PyFloat_AsDouble(PyList_GetItem(colList, 2));
	}
	return 0;
}


static PyObject* improve_calcPoint(PyObject *self, PyObject *args) {
	// Parse input
	PyObject *inMatrix, *inPoint;
	if (! PyArg_ParseTuple( args, "O!O!", &PyList_Type, &inMatrix, &PyList_Type, &inPoint)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}
	
	matrix model;
	if (getModelFromPython(inMatrix, model) != 0) return NULL;
	
	vec inVec, outVec;
	int i;
	for (i=0; i<3; i++){
		// point
		inVec[i] = PyFloat_AsDouble(PyList_GetItem(inPoint, i));
	}

	// invert
	calcPoint(model, inVec, outVec);

	// parse output
	PyObject *outPoint = PyList_New(0);
	for (i=0; i<3; i++) {
		PyList_Append(outPoint, PyFloat_FromDouble(outVec[i]));
	}
	return outPoint;
}



static PyObject* improve_invertMatrix(PyObject *self, PyObject *args) {
	// Parse input
	PyObject *inMatrix;
	if (! PyArg_ParseTuple( args, "O!", &PyList_Type, &inMatrix)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}
	
	int i;
	matrix model, inverse;
	for (i=0; i<3; i++){
		PyObject *colList = PyList_GetItem(inMatrix, i);
		if (PyList_Size(colList) != 3) {
			PyErr_SetString(PyExc_RuntimeError, "Matrix dimension failure");
			return NULL;
		}
		
		model[i][0] = PyFloat_AsDouble(PyList_GetItem(colList, 0));
		model[i][1] = PyFloat_AsDouble(PyList_GetItem(colList, 1));
		model[i][2] = PyFloat_AsDouble(PyList_GetItem(colList, 2));
	}

	// invert
	gslToMatrix(gsl_invert(matrixToGsl(model)), inverse);

	// parse output
	PyObject *list = PyList_New(0);
	int k;
	for (i=0; i<3; i++) {
		PyObject *sublist = PyList_New(0);
		for(k=0; k<3; k++) {
			PyList_Append(sublist, PyFloat_FromDouble(inverse[i][k]));
		}
		//printf("\n");
		PyList_Append(list, sublist);
	}
	return list;
}

static PyObject* improve_getTransPoints(PyObject *self, PyObject *args) {
	
	// get params
	PyObject *mList, *dTuple;
	if (! PyArg_ParseTuple( args, "O!O!", &PyList_Type, &mList, &PyTuple_Type, &dTuple)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}

	// dimensions
	int rwidth = (int)PyInt_AsLong(PyTuple_GetItem(dTuple, 0));
	int rheight = (int)PyInt_AsLong(PyTuple_GetItem(dTuple, 1));

	// model
	matrix model;
	if (getModelFromPython(mList, model) == 1) return NULL;
	
	// 1. Calc model points for 0/0, x/0, 0/y, x/y
	vec p1, p2, p3, p4;
	getPoint(model, 0, 0, p1);		// 0, 0
	getPoint(model, rwidth, 0, p2);		// x, 0
	getPoint(model, 0, rheight, p3);		// 0, y
	getPoint(model, rwidth, rheight, p4);	// x, y
	
	// 2. w' = (|f(0, 0) - f(x, 0)| + |f(x, y) - f(0, y)|) / 2
	// 3. h' = (|f(0, 0) - f(0, y)| + |f(x, y) - f(x, 0)|) / 2
	double wprime, hprime;
	vec p12, p13, p43, p42;
	vecDiff(p1, p2, p12); vecDiff(p1, p3, p13);
	vecDiff(p4, p3, p43); vecDiff(p4, p2, p42);
	// Min
	//wprime = min(vecNorm(p12, p43));
	//hprime = min(vecNorm(p13, p42));
	// Mean
	wprime = 0.5 * (vecNorm(p12) + vecNorm(p43));
	hprime = 0.5 * (vecNorm(p13) + vecNorm(p42));
	// Max
	//wprime = max(vecNorm(p12, p43));
	//hprime = max(vecNorm(p13, p42));
	
	// 4. adjust w', h' so that w'/h' = rwidth / rheight
	if (pow(wprime, 2) * ((double)rheight / (double)rwidth) > pow(hprime, 2) * ((double)rwidth / (double)rheight)) {
		hprime = wprime * rheight / rwidth;
	} else {
		wprime = hprime * rwidth / rheight;
	}
	
	// 4. Return 4 Points:	f(0, 0) <-> 0 , 0
	//			f(x, 0) <-> w', 0
	//			f(0, y) <-> 0 , h'
	//			f(x, y) <-> w', h'
	PyObject *finPoints = PyList_New(4);
	PyList_SetItem(finPoints, 0, buildPoint(p1[0], p1[1], +0.0, +0.0));
	PyList_SetItem(finPoints, 1, buildPoint(p2[0], p2[1], wprime, +0.0));
	PyList_SetItem(finPoints, 2, buildPoint(p3[0], p3[1], +0.0, hprime));
	PyList_SetItem(finPoints, 3, buildPoint(p4[0], p4[1], wprime, hprime));
	
	// 5. Return scale
	PyObject *scale = PyTuple_New(2);
	PyTuple_SetItem(scale, 0, PyFloat_FromDouble(wprime / (double)rwidth));		// fx
	PyTuple_SetItem(scale, 1, PyFloat_FromDouble(hprime / (double)rheight));	// fy
	
	PyObject *finRes = PyTuple_New(2);
	PyTuple_SetItem(finRes, 0, scale);
	PyTuple_SetItem(finRes, 1, finPoints);
	return finRes;
}

static PyObject* improve_getOrigin(PyObject *self, PyObject *args) {

	// get params
	PyObject *mList, *fTuple, *iTuple;
	if (! PyArg_ParseTuple( args, "O!O!O!", &PyList_Type, &mList, &PyTuple_Type, &fTuple, &PyTuple_Type, &iTuple)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}

	// dimensions reference
	int fx = (int)PyInt_AsLong(PyTuple_GetItem(fTuple, 0));
	int fy = (int)PyInt_AsLong(PyTuple_GetItem(fTuple, 1));

	// dimensions image
	int iwidth = (int)PyInt_AsLong(PyTuple_GetItem(iTuple, 0));
	int iheight = (int)PyInt_AsLong(PyTuple_GetItem(iTuple, 1));

	// model
	matrix model;
	if (getModelFromPython(mList, model) == 1) return NULL;
	

	// 5. invert(model)
	matrix inverse;
	vec p1, p2, p3, p4;
	gslToMatrix(gsl_invert(matrixToGsl(model)), inverse);
	getPoint(inverse, 0, 0, p1);			// 0, 0
	getPoint(inverse, iwidth, 0, p2);		// x, 0
	getPoint(inverse, 0, iheight, p3);		// 0, y
	getPoint(inverse, iwidth, iheight, p4);		// x, y
	
	// 6. fixpkt = (abs(minX(f'(0,0), f'(x,0), f'(0,y), f'(x,y))), abs(minY(f'(0,0), f'(x,0), f'(0,y), f'(x,y))))
	double ox, oy;
	// fx = wprime / (double)rwidth; fy = hprime / (double)rheight;
	ox = fx * abs(min(min(p1[0], p2[0]), min(p3[0], p4[0])));
	oy = fy * abs(min(min(p1[1], p2[1]), min(p3[1], p4[1])));
	
	// 7. Output
	PyObject *finOrigin = PyTuple_New(2);
	PyTuple_SetItem(finOrigin, 0, PyFloat_FromDouble(ox));
	PyTuple_SetItem(finOrigin, 1, PyFloat_FromDouble(oy));
	return finOrigin;
}

// Arguments: src ([Nx2] or [Nx3] Matrix), dst ([Nx2] or [Nx3] Matrix)
// Arguments: [Nx(1xM, 1xM)]: [(src,dst)]; M = {2,3}
// Returns: Homography ([3x3]-Matrix)
static PyObject* improve_getHomography(PyObject *self, PyObject *args) {
	
	// Get the params
	PyObject *mList;
	if (! PyArg_ParseTuple( args, "O!", &PyList_Type, &mList)) {
		PyErr_SetString(PyExc_RuntimeError, "Argument parse failure");
		return NULL;
	}

	// init lists
	int i, dim = PyList_Size(mList);
	
	CvMat* src = cvCreateMat(dim, 3, CV_32FC1);
	CvMat* dst = cvCreateMat(dim, 3, CV_32FC1);
	for (i=0; i<dim; i++){
		PyObject *srcList, *dstList, *tuple;
		
		// init srcList, dstList
		tuple = PyList_GetItem(mList, i);
		srcList = PyTuple_GetItem(tuple, 0);
		dstList = PyTuple_GetItem(tuple, 1);
		
		
		cvmSet(src, i, 0, PyFloat_AsDouble(PyList_GetItem(srcList, 0)));
		cvmSet(src, i, 1, PyFloat_AsDouble(PyList_GetItem(srcList, 1)));
		if (PyList_Size(srcList) >= 3) {
			cvmSet(src, i, 2, PyFloat_AsDouble(PyList_GetItem(srcList, 2))); // only, if given
		} else {
			cvmSet(src, i, 2, 1.0);
		}

		cvmSet(dst, i, 0, PyFloat_AsDouble(PyList_GetItem(dstList, 0)));
		cvmSet(dst, i, 1, PyFloat_AsDouble(PyList_GetItem(dstList, 1)));
		if (PyList_Size(dstList) >= 3) {
			cvmSet(dst, i, 2, PyFloat_AsDouble(PyList_GetItem(dstList, 2))); // only, if given
		} else {
			cvmSet(dst, i, 2, 1.0);
		}
		
	}

	// Calculate the homography
	CvMat* homography = cvCreateMat(3, 3, CV_32FC1);
	cvFindHomography(src, dst, homography, 0, 0, NULL);

	// return
	PyObject *list = PyList_New(0);
	int k;
	for (i=0; i<3; i++) {
		PyObject *sublist = PyList_New(0);
		for(k=0; k<3; k++) {
			PyList_Append(sublist, PyFloat_FromDouble(cvmGet(homography, i, k)));
		}
		PyList_Append(list, sublist);
	}
	
	cvReleaseMat(&src); cvReleaseMat(&dst);
	return list;
}



///*

static PyMethodDef RansacMethods[] = {
	{"ransac",  improve_ransac, METH_VARARGS, "Some temp module"},
	{"transform",  improve_calcPoint, METH_VARARGS, "Some temp module"},
	{"invert",  improve_invertMatrix, METH_VARARGS, "Some temp module"},
	{"getTransformPoints",  improve_getTransPoints, METH_VARARGS, "Some temp module"},
	{"getOrigin",  improve_getOrigin, METH_VARARGS, "Some temp module"},
	{"getHomography",  improve_getHomography, METH_VARARGS, "Some temp module"},
	{NULL, NULL, 0, NULL}        // Sentinel
};


PyMODINIT_FUNC initmodRansac(void) {
	(void) Py_InitModule("modRansac", RansacMethods);
}

// */


