
#include <math.h>
#include <stdio.h>

#include "ransac.h"


/**
 * Tries different thresholds
 * 
 * Initial value
 * - wildly guessed
 * - given
 * - src/img based
 * 	- max dist btw src[i], img[i]
 * 	- mean dist btw src[i], img[i]
 * 	- min dist btw src[i], img[i]
 * 
 * Modulation
 * - Wild guessing
 * - increase / decrease
 * 	- absolute factor		dst = dst +,- k;
 * 	- multiplicative factor		dst = dst *,/ k;
 * 	- statistically			dst *= k,
 * 					where k is the standart derivation of all distances
 * 		- 
 * 
 * Looping
 * - Break, if no improvement
 * - Best of m
 */

t_stats data_dist_stats(r_data *data) {
	double dst;
	int i;
	t_stats stats;
	
	stats.sum = 0;
	stats.max = 0;
	stats.var = 0;
	stats.min = euclid_dist(data->src[0], data->img[0]);
	
	for(i=0; i<data->dim; i++) {
		dst = euclid_dist(data->src[i], data->img[i]);
		
		stats.sum += dst;
		stats.var += pow(dst, 2);
		if (dst > stats.max) stats.max = dst;
		if (dst < stats.min) stats.min = dst;
	}
	
	stats.mean = stats.sum / data->dim;
	stats.var /= data->dim;
	stats.var -= pow(stats.mean, 2);
	return stats;
}


t_stats curve_stats(t_curve curve[], int length) {
	int i;
	t_stats stats;
	
	stats.sum = 0;
	stats.max = 0;
	stats.var = 0;
	stats.min = curve[0].y;
	
	for(i=0; i<length; i++) {
		stats.sum += curve[i].y;
		stats.var += pow(curve[i].y, 2);
		if (curve[i].y > stats.max) stats.max = curve[i].y;
		if (curve[i].y < stats.min) stats.min = curve[i].y;
	}
	
	stats.mean = stats.sum / length;
	stats.var /= length;
	stats.var -= pow(stats.mean, 2);
	return stats;
}

/*
	- Threshold	[TMin..TMax], TMin > 0
	- Subset	[0..1]
	- Qualität	(0..1]
	
	Qualität = Subset / Threshold
	norm_Qualität(i) = Qualität(i) / QMax
	
	- Je kleiner der Threshold, desto besser
	- Je grösser das Subset, desto besser
	- Je grösser die Qualität, desto besser
	
	=> Maximiere Qualität
		2) Lokales Maximum
		   - Bilde Qualität für 3 Punkte
		   - Gehe weiter in Richtung positiver Ableitung
		3) Globales Maximum
		   - Approximiere Daten mit Funktion (Polynom)
		   - Suche globales Maximum in Funktion (Ableiten, Nullstellensuche, Interpretation)
		4) Dynamisches, lokales Maximum
		   - Starte mit kleinstmöglichen Threshold (0)
		   - Bilde Qualität für n benachbarte Punkte
		   - Bilde Steigung ('lokale Ableitung')
		   - Wiederhole Verfahren weiter rechts, wenn Steigung nicht negativ (nicht unter bisherigen Mittelwert fällt)
		   - Suche Maximum aus gefundenen Punkten
	*/

#define APPROX_DIST_MEAN


/**
 * Determines a local maximum of the threshold quality 
 * 
 * It is not necessary to know how many steps are required to find the maximum
 * The algorithm assumes that the function will increase more or less monotonic
 * It stops, as soon as the gradient becomes negative
 * (with APPROX_DIST_MEAN it stops, as soon as the quality falls under the mean of
 * all qualities so far)
 * 
 * @param data		ransac data
 * @param params	ransac parameter. The dist will be changed !
 * @param mp_cnt	model points
 * @param step		step width
 * @param width	mean building width (class width)
 * @param chances	second chanse (n chances to recover)
 * 
 * @return		threshold (x) and quality (y) at maximum quality (not normalized)
 */
t_curve approx_dist_dyn_max(r_data *data, r_param *params, int mp_cnt, double step, int width, int chances) {
	
	const int maxloop = 1000;
	
	double r_val;
	matrix model;
	double m1, m2, qual;
	int i, lpcnt;
	t_curve max;
	
	#ifdef APPROX_DIST_MEAN
	double tot = 0;
	#endif
	
	max.x = 0; max.y = 0; m1 = 0;
	for(lpcnt=0; lpcnt < maxloop; lpcnt++) {
		// get mean quality of <width> points
		m2 = 0;
		for(i=0; i<width; i++) {
			r_val = ransac(data, params, mp_cnt, model);
			
			if (r_val <= 1 && r_val >= 0) {
				qual = r_val / params->dist;;
				m2 += qual;
			} else {
				qual = 0;
			}
			
			// get the maximum
			if (qual > max.y) {
				max.y = qual;
				max.x = params->dist;
			}
			
			params->dist += step;
		}
		m2 /= width;
		
		if (m2 < m1) {
			// gradient is negative
			// respectively m2 falls under the mean of all points so far
			if (chances > 0) {
				chances--;
			} else {
				break;
			}
		}
		
		#ifdef APPROX_DIST_MEAN
			tot += m2;
			m1 = tot / lpcnt;
		#else
			m1 = m2;
		#endif
	}
	
	params->dist = max.x;
	return max;
}


/**
 * Adjusts the threshold distance to a given set size
 * 
 */
/*int approx_dist_scurve() {
	return 0;
}*/


/**
 * Computes the quality of some thresholds
 * 
 * Computes the quality for each threshold (from params->dist to max) in <step> steps
 * The maximum will be calculated and returned
 * The threshold/quality tuples are stored in curve
 * (with APPROX_DIST_NORM the quality values are normalized)
 * 
 * @param data		ransac data
 * @param params	ransac parameter. The dist will be changed !
 * @param mp_cnt	model points
 * @param max		count of steps
 * @param step		step width
 * @param curve	the curve is stored here
 * 
 * @return		index of the curve element with the maximum quality
 */
int approx_dist_qcurve(r_data *data, r_param *params, int mp_cnt, int max, double step, t_curve curve[]) {
	
	double r_val, q_max=0;
	matrix model;
	int i, max_index=0;
	
	// initialization
	if (params->dist == 0) {
		params->dist = step;
	}
	
	// examine data
	for(i=0; i<max; i++) {
		r_val = ransac(data, params, mp_cnt, model);
		
		// calculate quality
		if (r_val <= 1 && r_val >= 0) {
			curve[i].y = r_val / params->dist;
		} else {
			curve[i].y = 0;
		}
		curve[i].x = params->dist;
		
		// one step forward
		params->dist += step;
		
		// maximum
		if (curve[i].y > q_max) {
			q_max = curve[i].y;
			max_index = i;
		}
	}
	
	#ifdef APPROX_DIST_NORM
	// normalization
	for(i=0; i<max; i++) {
		curve[i].y /= q_max;
	}
	#endif
	
	return max_index;
}

/**
 * 
 * 
 */
int approx_set_curve(r_data *data, r_param *params, int mp_cnt, double e, int max, double step, t_curve curve[]) {
	
	const int loopmax = 100;
	
	double r_val=0, s_max=0;
	matrix model;
	int i, max_index=0;
	
	// examine data
	for(i=0; i<max && e<1; i++) {
		//printf("%i (%f)\t", i, e);
		
		calcParams(params, e, data->dim, mp_cnt);
		if (params->loops > loopmax) params->loops = loopmax;
		if (params->set == 0) break;
		
		r_val = ransac(data, params, mp_cnt, model);
		
		// store in curve
		curve[i].x = params->set;
		curve[i].y = r_val;
		
		//printf("[%f %f] ... ", curve[i].x, curve[i].y);
		// maximum
		if (curve[i].y > s_max) {
			s_max = curve[i].y;
			max_index = i;
			//printf("MAX");
		}
		
		//printf("\n");
		// one step forward
		e += step;
	}
	
	for(; i<max; i++) {
		curve[i].x = params->set;
		curve[i].y = r_val;
	}
	
	return max_index;
}

/**
 * 
 * 
 */
int approx_loop_curve(r_data *data, r_param *params, int mp_cnt, double e, int max, double step, t_curve curve[]) {
	
	const int loopmax = 100;
	
	double r_val=0, l_max=0;
	matrix model;
	int i, max_index=0;
	
	// examine data
	for(i=0; i<max && e<1; i++) {
		
		//printf("%i (%f)\t", i, e);
		calcParams(params, e, data->dim, mp_cnt);
		if (params->set == 0) break;
		if (params->loops > loopmax) break;
		
		r_val = ransac(data, params, mp_cnt, model);
		
		// store in curve
		curve[i].x = params->loops;
		curve[i].y = r_val;
		
		//printf("[%f %f] ... ", curve[i].x, curve[i].y);
		// maximum
		if (curve[i].y > l_max) {
			l_max = curve[i].y;
			max_index = i;
			//printf("MAX");
		}
		
		//printf("\n");
		// one step forward
		e += step;
	}
	
	for(; i<max; i++) {
		curve[i].x = params->loops;
		curve[i].y = r_val;
	}
	
	return max_index;
}

/**
 * Estimates set size threshold and loop count
 * 
 * @param data		
 * @param params	
 * @param mp_cnt	
 * @param model_final	
 * @return		
 */
bool approx_set_and_loop(r_data *data, r_param *params, int mp_cnt) {
	
	const int maxloops = 10;
	const double mod = 0.25;
	
	int i;
	double set_cnt, e = 0.5, last_e;
	matrix model;
	
	// Initialization
	matrixSetIdentity(model);
	
	for(i=0; i<maxloops; i++) {
		
		if (e == 0) e=0.1;	/* FIXME */
		
		// rotate last values
		last_e = e;
		
		// Compute parameter out of e
		calcParams(params, e, data->dim, mp_cnt);
		
		// Simulate RANSAC
		printf("Computing with e = %f, set size = %i and loops = %i\n", e, params->set, params->loops);
		set_cnt = ransac(data, params, mp_cnt, model);
		printf("Return value is %f\t=> ", set_cnt);
		
		if (set_cnt < 0) {
			
			// no model - reduce error probability
			printf("Increasing e from %f", e);
			e += e * mod;
			//if (e == 0) e = mod;
			printf(" to %f\n", e);
		} else if (set_cnt > 1) {
			
			// several models - increase error probability
			printf("Decreasing e from %f", e);
			e -= e * mod;
			//if (e < 0) e = 0;
			printf(" to %f\n", e);
		} else {
			
			// one model
			if (1-set_cnt == e) {
				
				// Compute parameter out of e
				printf("No improvement for 1-set_cnt = %f and e = %f. Aborting\n", 1-set_cnt, e);
				calcParams(params, e, data->dim, mp_cnt);
				return TRUE;
			} else if (1-set_cnt < e) {
				
				printf("Not good enough, set e from %f to %f\n", e, 1-set_cnt);
				e = 1 - set_cnt;
			} else {
				
				printf("Got worse for 1-set_cnt = %f and e = %f. Aborting, Taking the last as result\n", 1-set_cnt, e);
				if (i > 0) {
					e = last_e;
					calcParams(params, e, data->dim, mp_cnt);
				}
				break;
			}
		}
	}
	
	
	if (set_cnt >= 0 && set_cnt <= 1) {
		return TRUE;
	} else {
		return FALSE;
	}
}



void calcModelStats(matrix model, r_data *data, r_param *param) {
	
}


q_prec calcMeanDist(matrix model, r_data *data) {
	int i;
	double distance = 0;
	vec cpoint;
	
	for(i=0; i<data->dim; i++) {
		// calculate model point
		calcPoint(model, data->src[i], cpoint);
		
		// calculate distance
		distance += euclid_dist(data->img[i], cpoint);
		
	}
	
	return distance / (double)(data->dim);
}


t_quality calcQuality(matrix model, r_data *data, r_param *params) {
	int i, set_cnt=0;
	double distance;
	t_quality quality;
	vec cpoint;
	
	// Initialization
	quality.tot = 0;
	quality.out = 0;
	quality.set = 0;
	
	for(i=0; i<data->dim ; i++) {
		// calculate model point
		calcPoint(model, data->src[i], cpoint);
		
		// calculate distance
		distance = euclid_dist(data->img[i], cpoint);
		
		// Add distance
		quality.tot += distance;
		if (distance < params->dist) {
			quality.set += distance;
			set_cnt++;
		}
	}
	
	// Build mean
	if (set_cnt != data->dim) {
		quality.out = (quality.tot - quality.set) / (double)(data->dim - set_cnt);
	} else {
		quality.out = 0;
	}
	
	if (set_cnt != 0) {
		quality.set /= (double)set_cnt;
	} else {
		quality.set = 0;
	}
	quality.tot /= (double)data->dim;
	return quality;
}

void calcParams(r_param *params, double e, int n, int m) {
	const double p = 0.99;
	
	params->set = ceil((1 - e) * n);
	params->loops = ceil( log( 1 - p ) / log( 1-pow(1-e, m) ) );
	params->e = e;
}


void evalModel(r_data *data, r_param *params, matrix model, double m_cnt, bool stat, FILE* output) {
	t_quality quality = calcQuality(model, data, params);
	double mdist = calcMeanDist(model,data);
	double distance;
	vec cpoint;
	int i;
	
	fprintf(output, "\nModel evaluation\n");
	for(i=0; i<data->dim; i++) {
		// calculate model point
		calcPoint(model, data->src[i], cpoint);
		
		// calculate distance
		distance = euclid_dist(data->img[i], cpoint);
		
		printVector(data->src[i], "SRC", FALSE, output);
		fprintf(output, "\t => \t");
		printVector(cpoint, "IMG", FALSE, output);
		fprintf(output, "\tind. Dist = %f\n", distance);
	}
	
	if (stat == TRUE) {
		fprintf(output, "Quality total\t= %f\nQuality set\t= %f\nQuality out\t= %f\nMean distance\t= %f\nThreshold quality = %f\n", quality.tot, quality.set, quality.out, mdist, m_cnt / params->dist);
	}
}

