
/**
 * 
 * FLAGS
 * - CSE_QUAL	Choose ouf of several models by its quality
 * - CSE_RAND	Choose out of several models randomly
 * 
 * - QUAL_SET	Base quality on set vectors
 * - QUAL_OUT	Base quality on outliers
 * (By default, the quality is based on all data)
 * 
 * - CKG_SUBSET	Computes the model only once per subset
 * - CKG_MODEL	Stores equal models only once
 * 
 * - PREEMPTIVE	preemptive RANSAC
 * 
 * - VERBOSE	Print some status information
 * - DEBUG	Print some more detailed information
 * 
 * 
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ransac.h"
#include "model.h"


/**
 * src		Original vectors
 * img		Image vectors
 * dim_data	Elements count
 * mp_cnt	Required model point count
 * loops	Loops
 * threshold_dst Distance threshold
 * threshold_set Minimum set size threshold
 * model_final	final model matrix
 * 
 * @return	-1	no model was found
 * 		[0..1]	one model was found, return the fraction of subset size over data dimension
 * 		> 1	several models were found
 * 
 * @todo	Avoid choice of the same basis two times
 * @todo	Repeat subset build with smaller threshold
 * @todo	Use different random number generator (true random generator or randomized seed)
 */
double ransac(r_data *data, r_param *params, int mp_cnt, matrix model_final) {

	/**
	 * LOCALS
	 */
	
	struct {
		int set[data->dim];
		int len;
	} set_list[params->loops], *set_current, *set_max;			// set list, current set, maximum size set
	
	matrix model;								// model matrix
	vec basis_src[mp_cnt], basis_img[mp_cnt], cpoint;			// basis, calculated image point
	double distance=0;							// distance
	bool found;								// element found flag

	int mod_cnt=0, lp_cnt = params->loops, max_cnt = 1;		 	// model count, loop count, maximum subset count
	int used[mp_cnt];							// used src indices
	int size_int = sizeof(int), i, rval;					// size of int, counter, random value

	
	/**
	 * INITIALIZATION
	 */
	
	// init random generator
	srand(876543);
	
	
	/**
	 * RANSAC
	 */
	
	set_max = &(set_list[lp_cnt - 1]);
	for(lp_cnt = params->loops; lp_cnt > 0; lp_cnt--) {
		// prepare set
		set_current = &(set_list[lp_cnt - 1]);
		set_current->len = 0;
		
		for(i=0;i<mp_cnt;i++) {
			used[i] = -1;
		}
		
		// determine mp_cnt points, not already used
		do {
			
			// get mp_cnt points
			i = 0;
			do {
				
				rval = rand() % data->dim;
				if (in_array(used, mp_cnt, size_int, &rval) < 0) {
					
					used[i] = rval;
					copyVec(basis_src[i], data->src[rval]);
					copyVec(basis_img[i], data->img[rval]);
					i++;
				}
			} while (i < mp_cnt);
			
		} while(0);	// while (<ALREADY USED>)
		
		// determine model
		if (MODEL_FUNCTION(basis_src, basis_img, mp_cnt, model) == FALSE) {
			#ifdef DEBUG
				printf("ERROR: Model cannot be determined from subset");
				printIntArray(used, mp_cnt, "");
			#endif
			continue;
		}
		
		//printf("\n");
		//printMatrix(model, "MODEL");
		// for each point
		for(i=0;i<data->dim;i++) {
			
			// calculate model point
			calcPoint(model, data->src[i], cpoint);
			
			// calculate distance
			distance = euclid_dist(data->img[i], cpoint);
			//printVector(cpoint, "POINT", FALSE);
			//printf("\tDistance = %f\n", distance);
			
			if (distance < params->dist) {
				
				// store in consensus set
				set_current->set[set_current->len] = i;
				// increment consensus set count
				set_current->len++;
			}
		}
		
		
		#ifdef PREEMPTIVE
		if (set_current->len >= params->set) {
			#ifdef VERBOSE
				printf("Preempt algorithm with set size %i (>= %i)\n", set_current->len, threshold_set);
			#endif
			setToModel(data, set_current->set, set_current->len, model_final);
			return set_current->len / data->dim;
		}
		#endif
		
		// maximum set
		if (set_current->len > set_max->len) {
			
			// reset max counter
			max_cnt = 1;
			
			// assign to max
			set_max = set_current;
			
		} else if (set_current->len == set_max->len && set_current != set_max) {
			
			// increment max counter
			max_cnt++;
		}
	}
	
	#ifdef DEBUG
		for(i=0; i<params->loops; i++) {
			printf("Subset %i", i);
			printIntArray(set_list[i].set, set_list[i].len, "");
		}
		printf("\n");
	#endif
	
	#ifdef VERBOSE
		if (max_cnt != 1) {
			printf("There exist %i subsets with maximum length (= %i)\n", max_cnt, set_max->len);
		} else {
			printf("There exist 1 subset with maximum length (= %i)\n", set_max->len);
		}
	#endif
	
	
	
	/**
	 * CHOOSE MODEL
	 */
	
	
	/** SET SIZE CHECK */
	
	// check minimum consensus size
	if (set_max->len < params->set) {
		
		// no large enough consensus set has been found
		#ifdef VERBOSE
			printf("There exists no model with minimum set size (= %i > %i)\n", params->set, set_max->len);
		#endif
		matrixSetIdentity(model_final);
		return -1;
	}
	
	
	struct {
		matrix model;
		double quality;
	} set_models[max_cnt];
	
	
	/** MODEL CALCULATION */
	
	// calculate model for all consensus sets
	mod_cnt = 0;
	for(i=0; i<params->loops; i++) {
		
		// only look at elements with a maximum subset
		if (set_list[i].len == set_max->len) {
			
			// check, if the model for this subset has already been calculated
			found = FALSE;
			
			#ifdef CKG_SUBSET
			int k;
			for(k=i-1; k > 0; k--) {
				
				if (set_max->len == set_list[k].len) {
					
					// compare sets
					if (cmpArray(set_list[i].set, set_list[k].set, size_int, set_list[i].len) == TRUE) {
						
						found = TRUE;
						break;
					}
				}
			}
			#endif
			
			if (found == FALSE) {
				
				// calculate model
				if (setToModel(data, set_list[i].set, set_list[i].len, model) == TRUE) {
				
					#ifdef CKG_MODEL
					int k;
					// check, if another subset evaluated to the identical model
					for(k=0; k < mod_cnt; k++) {
						
						// compare models
						if (cmpArray(set_models[k].model, model, sizeof(c_prec), DIM * DIM) == TRUE) {
							
							found = TRUE;
							break;
						}
					}
					#endif
					
					if (found == FALSE) {
						#ifdef DEBUG
							printMatrix(model, "Model", stdout);
						#endif
						
						// append onto model list - the model is unique in the array!
						copyMatrix(set_models[mod_cnt].model, model);
						mod_cnt++;
					}
				#ifdef DEBUG
				} else {
					printf("ERROR: Model cannot be determined from subset ");
					printIntArray(set_list[i].set, set_list[i].len, "");
				#endif
				}
			}
		}
	}
	
	#ifdef VERBOSE
		if (mod_cnt != 1) {
			printf("There remain %i models\n", mod_cnt);
		} else {
			printf("There remains 1 model\n");
		}
	#endif
	
	#ifdef DEBUG
	if (mod_cnt < 1) {
		printf("ERROR: No consensus subset available (mod_cnt < 1)\n");
		exit(0);
	}
	#endif
	
	/** QUALITY CHECK */
	
	#ifdef CSE_QUAL
	// check, if models need to be sorted by quality
	if (mod_cnt > 1) {
		
		#ifdef VERBOSE
			printf("Reduced %i models to", mod_cnt);
		#endif
		
		q_prec min_error = 1000000;
		
		// calculate qualities
		t_quality tmp_quality;
		for(i=0; i < mod_cnt; i++) {
			
			tmp_quality = calcQuality(set_models[i].model, data, params);
			#ifdef QUAL_SET
				set_models[i].quality = tmp_quality.set;
			#elif QUAL_OUT
				set_models[i].quality = tmp_quality.out;
			#else
				set_models[i].quality = tmp_quality.tot;
			#endif
			if (set_models[i].quality < min_error) {
				min_error = set_models[i].quality;
			}
		}
		
		// sort bad ones out
		for(i=0; i < mod_cnt; i++) {
			
			if (set_models[i].quality > min_error) {
				set_models[i] = set_models[mod_cnt-1];
				mod_cnt--;
			}
		}
		
		#ifdef VERBOSE
			printf(" %i (Quality = %f)\n", mod_cnt, min_error);
		#endif
	}
	#endif
	
	#ifdef DEBUG
	if (mod_cnt < 1) {
		printf("ERROR: No consensus subset available (mod_cnt < 1)\n");
		exit(0);
	}
	#endif
	
	/** RANDOM CHOICE */
	
	#ifdef CSE_RAND
	// check, if model need to be choosen by random
	if (mod_cnt > 1) {
		
		// choose a random model
		rval = rand() % mod_cnt;
		set_models[0] = set_models[rval];
		
		#ifdef VERBOSE
		if (rval == 0) {
			printf("Choose the 1st subset out of %i randomly\n", mod_cnt);
		} else if (rval == 1) {
			printf("Choose the 2nd subset out of %i randomly\n", mod_cnt);
		} else if (rval == 2) {
			printf("Choose the 3rd subset out of %i randomly\n", mod_cnt);
		} else {
			printf("Choose the %ith subset out of %i randomly\n", rval+1, mod_cnt);
		}
		#endif
		
	}
	#endif
	
	/** TERMINATION */
	
	// copy the final model
	copyMatrix(model_final, set_models[0].model);
	
	// return model count
	if (mod_cnt < 1) {
		return -1;
	} else if (mod_cnt > 1) {
		return mod_cnt;
	} else {
		return (double)set_max->len / (double)(data->dim);
	}
}


/**
 * Calcuates a model out of a set
 * 
 * @param data		input vectors
 * @param set		set array
 * @param len		set length
 * @param model		Target model
 * @return		return value of MODEL_FUNCTION
 */
bool setToModel(r_data *data, int set[], int len, matrix model) {
	int i;
	vec basis_src[len];
	vec basis_img[len];
	
	for(i=0; i<len; i++) {
		copyVec(basis_src[i], data->src[set[i]]);
		copyVec(basis_img[i], data->img[set[i]]);
	}
	
	// calculate model
	return MODEL_FUNCTION(basis_src, basis_img, len, model);
}


